#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "GraphicsCom.c"
#include "Graf.h"
#include "grafprot.h"

/* This function draws text in the current Laser Writer font.  Parameters:

x,y:          Coordinates of text

theta:        Angle of display (degrees)

coord_code:   Indicates type of coordinates; 0=user, 1=fiducial, 2=inch,
              the latter measured relative to the LL corner of the page.

lr_center:    Indicates horizontal centering: 0=left, 1=center, 2=right.

bt_center:    Indicates vertical centering:  0=bottom, 1=center.

blank_bkgnd:  If True, indicates that background should be blanked out
              before text is written out.

text:         The text to be written, null terminated.  If the text is
              to be displayed in a LaserWriter (PostScript) font, it 
              must follow the PostScript rules for string constants
              (see p.29, PostScript Language Reference, 3rd ed).  If
              it is to be displayed in TeX fonts, it is assumed that TeX
              is in horizontal mode when the string is encountered.  Thus,
              if you want math mode, you have to surround the string
              with $ signs.  */

void DrawText(double x,double y,double theta,int coord_code,
	      int lr_center,int bt_center,Bool blank_bkgnd,char *text)
{
  int iysh,ix,iy,margin,marg2;

  /* Check parameters for validity */

  if(coord_code<0 || coord_code>2 || lr_center<0 || lr_center>2
     || bt_center<0 || bt_center>1){
    fprintf(stderr,"Invalid parameters in DrawText\n");
    return;
  }

  /* If inch coordinates requested, convert them to fiducial coords
     and use them subsequently.  The rest of the function needs to
     deal only with user,fiducial coords */

  if(coord_code==2){
    x=(x-G.xfid)/G.fidconvi;
    y=(y-G.yfid)/G.fidconvi;
    coord_code=1;
  }

  /* Make sure proper font exists */

  if(G.icurfnt<0){
    fprintf(stderr,"Function DrawText: No font specified\n");
    return;
  }

  /* Convert x,y specified by caller into ix,iy, pixel coordinates
     of the reference point.  */

  if(coord_code==0){
    ix=UserxToPix(x);
    iy=UseryToPix(y);
  }
  else{
    ix=nint(G.conv*(G.xfid+x*G.fidconvi));
    iy=nint(G.conv*(G.yfid+y*G.fidconvi));
  }

  /* What follows is a stripped down version of postscript generated 
     by old dos txtpass5, assuming that we have only a simple string
     to display. */

  fprintf(G.psfile,"/str0 (%s) def\n",text);
  fprintf(G.psfile,"/gw000 0 def /gh000 0 def /gd000 0 def 0 0 moveto\n");
  fprintf(G.psfile,"str0 strhd\n");
  fprintf(G.psfile,"dup gh000 gt {/gh000 exch def} {pop} ifelse\n");
  fprintf(G.psfile,"dup gd000 lt {/gd000 exch def} {pop} ifelse\n");
  fprintf(G.psfile,"str0 stringwidth rmoveto\n");
  fprintf(G.psfile,"currentpoint pop dup gw000 gt {/gw000 exch def}"
                   " {pop} ifelse\n");

  /* Generate postscript in preparation for displaying text.  First
     rotate, translate origin, to place origin at reference point of
     text.  Then advance currentpoint to lower left corner of
     text to be displayed */

  fprintf(G.psfile,"%d %d translate %8.2f rotate\n",ix,iy,theta);
  iysh=-G.ifontsz*bt_center/3;
  fprintf(G.psfile,"gw000 %d mul 2 div neg %d moveto\n",lr_center,iysh);

  /* Now generate blanking box, if called for */

  if(blank_bkgnd){
    margin=nint(0.15*(double)G.ifontsz);
    marg2=2*margin;
    fprintf(G.psfile,"currentpoint %d gd000 %d sub rmoveto\n",-margin,margin);
    fprintf(G.psfile,"currentpoint newpath moveto gw000 %d add 0 rlineto\n",
	    marg2);
    fprintf(G.psfile,"0 gh000 gd000 sub %d add rlineto gw000 %d add\n",
	    marg2,marg2);
    fprintf(G.psfile,"neg 0 rlineto closepath 1.0 setgray fill 0 "
                     "setgray moveto\n");
  }

  /* Now generate stuff from old txtpass6 */

  fprintf(G.psfile,"currentpoint\n");
  fprintf(G.psfile,"str0 show\n");
  fprintf(G.psfile,"moveto\n");
  fprintf(G.psfile,"gw000 0 rmoveto\n");

  /* Now restore postscript coordinates */

  fprintf(G.psfile,"%10.2f rotate %d %d translate\n",-theta,-ix,-iy);
    
}
