#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "wallprot.h"
#include "potprot.h"
#include "grafprot.h"

/* Same as freeexpand2 except a gravitational field is included. */

#define MIN(x,y) ((x)<(y)?(x):(y))
#define MAX(x,y) ((x)>(y)?(x):(y))

typedef struct particle particle;
struct particle{
  double x;
  double y;
  double px;
  double py;
};

/* On each iteration, we sort the particles by the box they belong to.
   The box structure describes a single box. 

   nparticles:         Number of particles in the box
   nneighbors:         Number of neighboring boxes; includes current
                       box itself; usually 9, but less on boundaries
   particle_indexes:   Array of indexes of particles that are in the
                       current box
   pneighbors;         Array of pointers to neighboring boxes
*/

typedef struct box box;
struct box{
  int nparticles;
  int nneighbors;
  int *particle_indexes;
  box **pneighbors;
};

#define Nparts 12000   /* Number of particles */

static void print_energy(int iframe, particle parts[], int nparts);
static void frameit(int iframe,particle parts[], int nparts);
static void message_frame(int iframe);
static void draw_thermometer();

static double a,asq,xright,ytop,b,xwall,g;
static int redpart,nmessageframes;

int main()
{
  int nsteps,i,nframes,iframe,istep,j;
  double tau,kT0,vx,vy,rootkT0;
  static particle parts[Nparts],parts1[Nparts];
  double x1,y1,dx,dy,drsq,dfx,dfy,xpnt,fx,fy,dr,rderiv,rwell;
  double px,py;
  int nxboxes,nyboxes;
  double range,f;
  box **boxes,**pneighbors,*pneighbor;
  int *particle_indexes;
  int k,kmin,kmax,m,mmin,mmax,dk,dm,nneighbors,n,nparticles;
  int ipart,jpart,nsecs;
  double gravenergy,kTtarget;
  FILE *statefile;

  /* a is distance parameter in potential energy functions. eg
     exp(-(r/a)^2).

     rwell is distance to attractive well, in units of a

     range is dist beyond which potential is negligable

     kT0 is the average initial kinetic energy.  Potential energy is
     extra.  Thus, this is not the actual temperature of the system,
     only a nominal temperature that is used for setting up initial
     conditions.  The actual temperature is printed out periodically.

     kTtarget is the desired temperature.  It is achieved by adjusting
     the kinetic energy of the particles, adding or subtracting
     energy, at each frame.

     tau is the time step of the integrator.

     nsteps is the number of time steps per frame

     nsecs is the duration of the movie, in real clock seconds

     nframes is the number of frames in the movie.  At 25 frames/sec,
     the number of seconds the movie lasts is nframes/25, or in
     minutes, nframes/1500.

     x bounds of box are [0,xright], y bounds are [0,ytop]

     redpart is the particle that is colored red
  */
     

  a=0.005;
  asq=a*a;
  rwell=2.0;
  range=(rwell+6.07)*a;
  init_ppot1(2.0,-0.5,rwell);
  kT0=0.4;
  kTtarget=0.5;
  rootkT0=sqrt(kT0);
  tau=0.000125;
  nsteps=64;
  nsecs=100;
  nframes=nsecs*25;
  xright=2.33;
  ytop=1.66;
  redpart=Nparts-1;
  b=0.5*a;
  xwall=xright;

  /* Get acceleration of gravity.  Make the energy to climb from
     bottom of box to top gravenergy.  */

  gravenergy=0.3;
  g=gravenergy/ytop;

  nmessageframes=2*25;
  
  for(iframe=0;iframe<nmessageframes;++iframe)
    message_frame(iframe);

  /* Set up box arrays. */
    
  nxboxes=ceil(xright/range);
  nyboxes=ceil(ytop/range);

  if((boxes=(box**)malloc(nxboxes*sizeof(box*)))==NULL){
    printf("Mem err #1\n");
    exit(1);
  }

  for(i=0;i<nxboxes;++i){
    if((boxes[i]=(box*)malloc(nyboxes*sizeof(box)))==NULL){
      printf("Mem err #2\n");
      exit(1);
    }
  }

  /* Initialize boxes.  Have to count the number of neighbors,
     set up pointers to neighboring boxes.  Also have to allocate
     arrays for particles in a box and for pointers to neighboring
     boxes. 
  */

  for(i=0;i<nxboxes;++i){
    kmin=MAX(i-1,0);
    kmax=MIN(i+1,nxboxes-1);
    dk=kmax-kmin+1;
    for(j=0;j<nyboxes;++j){
      boxes[i][j].nparticles=0;
      if((boxes[i][j].particle_indexes=(int*)malloc(Nparts*sizeof(int)))
	 ==NULL){
	printf("Mem err #3\n");
	exit(1);
      }

      mmin=MAX(j-1,0);
      mmax=MIN(j+1,nyboxes-1);
      dm=mmax-mmin+1;
      nneighbors=dk*dm;
      boxes[i][j].nneighbors=nneighbors;
      if((boxes[i][j].pneighbors=(box**)malloc(nneighbors*sizeof(box*)))
	 ==NULL){
	printf("Mem err #4\n");
      }

      n=0;
      for(k=kmin;k<=kmax;++k){
	for(m=mmin;m<=mmax;++m){
	  boxes[i][j].pneighbors[n]=&(boxes[k][m]);
	  ++n;
	}
      }
    }
  }

  /* Set up initial conditions */

  for(i=0;i<Nparts;++i){
    parts[i].x=xright*urand();
    if(g==0.0)
      parts[i].y=ytop*urand();
    else
      parts[i].y = -(kT0/g)*log(1.0-urand()*(1.0-exp(-g*ytop/kT0)));
    gaurand(&vx,&vy);
    parts[i].px=rootkT0*vx;
    parts[i].py=rootkT0*vy;
  }

  print_energy(0,parts,Nparts);
  frameit(nmessageframes,parts,Nparts);

  for(iframe=1;iframe<=nframes;++iframe){
    for(istep=0;istep<nsteps;++istep){

      /* Symplectic leapfrog integrator.  First advance positions half
	 step. */

      for(ipart=0;ipart<Nparts;++ipart){
	parts1[ipart].x=parts[ipart].x+0.5*tau*parts[ipart].px;
	parts1[ipart].y=parts[ipart].y+0.5*tau*parts[ipart].py;
      }

      /* Sort particles by boxes they belong to.  */

      for(i=0;i<nxboxes;++i){
	for(j=0;j<nyboxes;++j){
	  boxes[i][j].nparticles=0;
	}
      }

      for(ipart=0;ipart<Nparts;++ipart){
	x1=parts1[ipart].x;
	y1=parts1[ipart].y;
	j=floor(x1/range);
	j=MAX(j,0);
	j=MIN(j,nxboxes-1);
	k=floor(y1/range);
	k=MAX(k,0);
	k=MIN(k,nyboxes-1);   

	/* (x1,y1) belongs to box (j,k) */

	n=boxes[j][k].nparticles;
	boxes[j][k].particle_indexes[n]=ipart;
	boxes[j][k].nparticles=n+1;
      }

      /* Compute forces on particles at half-stepped position. */

      for(ipart=0;ipart<Nparts;++ipart){
	x1=parts1[ipart].x;
	y1=parts1[ipart].y;

	/* First forces from walls */

	fx=(-wf0(-x1/b)+wf0((x1-xwall)/b))/b;
	fy=(-wf0(-y1/b)+wf0((y1-ytop)/b))/b;

	/* Now forces from particles in neighboring boxes.  Others are
	   too far away to matter.  */

	dfx=0.0;
	dfy=0.0;

	i=floor(x1/range);
	i=MAX(i,0);
	i=MIN(i,nxboxes-1);
	j=floor(y1/range);
	j=MAX(j,0);
	j=MIN(j,nyboxes-1);  

	/* (x1,y1) belongs to box (i,j).   Add up forces from all
	   particles in all neighboring boxes, but don't include any
	   self-force. */

	nneighbors=boxes[i][j].nneighbors;
	pneighbors=boxes[i][j].pneighbors;
	for(k=0;k<nneighbors;++k){
	  pneighbor=pneighbors[k];
	  nparticles=pneighbor->nparticles;
	  particle_indexes=pneighbor->particle_indexes;
	  for(m=0;m<nparticles;++m){
	    jpart=particle_indexes[m];
	    if(jpart==ipart)continue;
	    dx=x1-parts1[jpart].x;
	    dy=y1-parts1[jpart].y;
	    drsq=dx*dx+dy*dy;
	    dr=sqrt(drsq);
	    rderiv=pv1deriv(dr/a);
	    dfx+=-dx*rderiv;
	    dfy+=-dy*rderiv;
	  }
	}
	fx+=dfx/asq;
	fy+=dfy/asq;

	/* Now gravitational force. */

	fy+=-g;

	/* All forces on particle ipart are now computed in (fx,fy).
	   Advance the momentum a full tau step.  */
	  
	parts1[ipart].px=parts[ipart].px+tau*fx;
	parts1[ipart].py=parts[ipart].py+tau*fy;
      }

      /* Now advance the position the remaining half step. */


      for(ipart=0;ipart<Nparts;++ipart){
	parts1[ipart].x+=0.5*tau*parts1[ipart].px;
	parts1[ipart].y+=0.5*tau*parts1[ipart].py;
      }

      /* Copy the advanced state back to original state */

      for(ipart=0;ipart<Nparts;++ipart)parts[ipart]=parts1[ipart];
    }

    /* nsteps intervals tau completed.  Generate a frame.  */

    frameit(iframe+nmessageframes,parts,Nparts);
    print_energy(iframe,parts,Nparts);
    if(iframe==125)xwall=xright;
  }

  /* Write the state of the system to file state.dat, so that you can
     resume simulation from this point if it looks good so far.  */

  if((statefile=fopen("state.dat","w"))==NULL){
    printf("Cannot open state.dat\n");
    exit(1);
  }
  
  fwrite(&nframes,sizeof(int),1,statefile);
  fwrite(parts,sizeof(particle),Nparts,statefile);
  fclose(statefile);

  exit(0);
}

static void print_energy(int iframe, particle parts[], int nparts)
{
  double E,px,py,x,y,xpnt,dx,dy,dr,ke,kT;
  int ipart,jpart;

  E=0.0;
  ke=0.0;
  for(ipart=0;ipart<nparts;++ipart){
    px=parts[ipart].px;
    py=parts[ipart].py;
    ke+=0.5*(px*px+py*py);
    x=parts[ipart].x;
    y=parts[ipart].y;
    E+=wv0(-x/b)+wv0((x-xwall)/b)+wv0(-y/b)+wv0((y-ytop)/b);
    for(jpart=0;jpart<ipart;++jpart){
      dx=x-parts[jpart].x;
      dy=y-parts[jpart].y;
      dr=sqrt(dx*dx+dy*dy);
      E+=pv1(dr/a);
    }
    E+=g*y;
  }
  E+=ke;
  kT=ke/Nparts;   /* Converts ke to temperature */

  printf("%4d kT=%10.5f, E=%10.5f\n",iframe,kT,E);
}

static void frameit(int iframe,particle parts[], int nparts)
{
  static char filename[20];
  static CurvePoint walls[5],barrier[5];
  int i;

  sprintf(filename,"frames/%04d.ps",iframe);
  SetGraphicsFileName(filename);
  if(OpenGraphics(15.0,11.0)==1){
    printf("Can't open graphics file %d\n",iframe);
    exit(1);
  }
  NewPage(0);
  SetFiducial(0.5,0.5,6.0);
  SetWindow(0.0,xright,0.0,ytop, 0,0.0,xright, 0,0.0,ytop);
  SetBezier(0);

  walls[0].x=0.0;
  walls[0].y=0.0;
  walls[1].x=xright;
  walls[1].y=0.0;
  walls[2].x=xright;
  walls[2].y=ytop;
  walls[3].x=0.0;
  walls[3].y=ytop;
  walls[4]=walls[0];

  barrier[0].x=0.4*xright;
  barrier[0].y=0.0;
  barrier[1].x=0.4*xright;
  barrier[1].y=ytop;
  barrier[2].x=0.4*xright+5.0*b;
  barrier[2].y=ytop;
  barrier[3].x=0.4*xright+5.0*b;
  barrier[3].y=0.0;
  barrier[4]=barrier[0];

  SetLineWidth(10);
  DrawCurve2(walls,5,True,True,True,True,False);
  /*  Old code used in free expansion program.
  if(iframe<=nmessageframes+125){
   DrawCurve2(barrier,5,True,True,True,False,True);
  }
  */

  /* Make red atom bigger, easier to see.  */

  for(i=0;i<nparts;++i){
    if(i!=redpart)DrawBullet(parts[i].x,parts[i].y,0.004);
    else{
      SetColor(1.0,0.0,0.0);   /* RGB values */
      DrawBullet(parts[i].x,parts[i].y,0.007);
      ResetGray();
    }
  }
  draw_thermometer();
  EndPage();
  CloseGraphics();
}

static void draw_thermometer()
{
  double xtleft,xtright,ytbottom,yttop;
  static CurvePoint bdry[5];

  xtleft=xright*0.95;
  xtright=xright;
  ytbottom=0.33*ytop;
  yttop=ytop;

  bdry[0].x=xtleft;
  bdry[0].y=ytbottom;
  bdry[1].x=xtright;
  bdry[1].y=ytbottom;
  bdry[2].x=xtright;
  bdry[2].y=yttop;
  bdry[3].x=xtleft;
  bdry[3].y=yttop;
  bdry[4]=bdry[0];

  SetGray(1.0);   /* White */
  DrawCurve2(bdry,5,True,True,True,False,True);
  SetGray(0.0);  /* Black again */
}

static void message_frame(int iframe)
{
  static char filename[20];

  sprintf(filename,"frames/%04d.ps",iframe);
  SetGraphicsFileName(filename);
  if(OpenGraphics(15.0,11.0)==1){
    printf("Can't open graphics file %d\n",iframe);
    exit(1);
  }
  NewPage(0);
  SetFiducial(0.5,0.5,6.0);
  SetWindow(0.0,xright,0.0,ytop, 0,0.0,xright, 0,0.0,ytop);
  SetBezier(0);
  SetLineWidth(5);
  SetFont(0,0.1);
  DrawText(7.5,7.0,0.0, 2,1,1, False,"Rep Core, Attrac Well + Grav");

  EndPage();
  CloseGraphics();
}
