#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "wallprot.h"
#include "grafprot.h"

#define MIN(x,y) ((x)<(y)?(x):(y))
#define MAX(x,y) ((x)>(y)?(x):(y))

typedef struct particle particle;
struct particle{
  double x;
  double y;
  double px;
  double py;
};

/* On each iteration, we sort the particles by the box they belong to.
   The box structure describes a single box. 

   nparticles:         Number of particles in the box
   nneighbors:         Number of neighboring boxes; includes current
                       box itself; usually 9, but less on boundaries
   particle_indexes:   Array of indexes of particles that are in the
                       current box
   pneighbors;         Array of pointers to neighboring boxes
*/

typedef struct box box;
struct box{
  int nparticles;
  int nneighbors;
  int *particle_indexes;
  box **pneighbors;
};

#define Nparts 4000   /* Number of particles */

static void print_energy(int iframe, particle parts[], int types[], 
			 int nparts);
static void frameit(int iframe,particle parts[], int types[], int nparts);
static void message_frame();

static double a,asq,B,xright,ytop,b;
static int frameno;
static int indexes[Nparts];

int main()
{
  int nsteps,i,nframes,iframe,istep,j;
  double tau,kT,vx,vy,rootkT;
  static particle parts[Nparts],parts1[Nparts];
  static int types[Nparts];
  double x1,y1,dx,dy,drsq,dfx,dfy,xpnt,fx,fy;
  double px,py;
  double exp1,exp2,expcombo;
  int nxboxes,nyboxes;
  double range,f;
  box **boxes,**pneighbors,*pneighbor;
  int *particle_indexes;
  int k,kmin,kmax,m,mmin,mmax,dk,dm,nneighbors,n,nparticles;
  int ipart,jpart,it;

  /* a is distance parameter in potential energy functions. eg
     exp(-(r/a)^2).

     range is dist beyond which Gaussian potential is negligable

     B is the coefficient of the interparticle potential energy

     kT is the average kinetic energy.  Potential energy is extra.

     tau is the time step of the integrator.

     nsteps is the number of time steps per frame

     nframes is the number of frames in the movie.  At 25 frames/sec,
     the number of seconds the movie lasts is nframes/25, or in
     minutes, nframes/1500.

     x bounds of box are [0,xright], y bounds are [0,ytop]

  */
     

  a=0.005;
  range=12.0*a;
  asq=a*a;
  B=2.0;
  kT=0.25;
  rootkT=sqrt(kT);
  tau=0.000125;
  nsteps=32;
  xright=2.33;
  ytop=1.66;
  b=0.5*a;

  frameno=0;
  nframes=0*25;
  for(iframe=0;iframe<nframes;++iframe){
    message_frame();
    ++frameno;
  }

  /* Get a random set of indexes */

  for(i=0;i<Nparts;++i)indexes[i]=i;
  for(i=0;i<10*Nparts;++i){
    ipart=urand()*Nparts;
    ipart=MAX(ipart,0);
    ipart=MIN(ipart,Nparts-1);
    jpart=urand()*Nparts;
    jpart=MAX(jpart,0);
    jpart=MIN(jpart,Nparts-1);
    it=indexes[ipart];
    indexes[ipart]=indexes[jpart];
    indexes[jpart]=it;
  }

  nframes=399*25;

  /* Set up box arrays. */

  nxboxes=ceil(xright/range);
  nyboxes=ceil(ytop/range);

  if((boxes=(box**)malloc(nxboxes*sizeof(box*)))==NULL){
    printf("Mem err #1\n");
    exit(1);
  }

  for(i=0;i<nxboxes;++i){
    if((boxes[i]=(box*)malloc(nyboxes*sizeof(box)))==NULL){
      printf("Mem err #2\n");
      exit(1);
    }
  }

  /* Initialize boxes.  Have to count the number of neighbors,
     set up pointers to neighboring boxes.  Also have to allocate
     arrays for particles in a box and for pointers to neighboring
     boxes. 
  */

  for(i=0;i<nxboxes;++i){
    kmin=MAX(i-1,0);
    kmax=MIN(i+1,nxboxes-1);
    dk=kmax-kmin+1;
    for(j=0;j<nyboxes;++j){
      boxes[i][j].nparticles=0;
      if((boxes[i][j].particle_indexes=(int*)malloc(Nparts*sizeof(int)))
	 ==NULL){
	printf("Mem err #3\n");
	exit(1);
      }

      mmin=MAX(j-1,0);
      mmax=MIN(j+1,nyboxes-1);
      dm=mmax-mmin+1;
      nneighbors=dk*dm;
      boxes[i][j].nneighbors=nneighbors;
      if((boxes[i][j].pneighbors=(box**)malloc(nneighbors*sizeof(box*)))
	 ==NULL){
	printf("Mem err #4\n");
      }

      n=0;
      for(k=kmin;k<=kmax;++k){
	for(m=mmin;m<=mmax;++m){
	  boxes[i][j].pneighbors[n]=&(boxes[k][m]);
	  ++n;
	}
      }
    }
  }

  /* Set up initial conditions */

  for(i=0;i<Nparts;++i){
    parts[i].x=xright*urand();
    parts[i].y=ytop*urand();
    gaurand(&vx,&vy);
    parts[i].px=rootkT*vx;
    parts[i].py=rootkT*vy;
    if(i<Nparts/2)types[i]=0;
    else types[i]=1;
  }

  print_energy(0,parts,types,Nparts);
  frameit(0,parts,types,Nparts);
  ++frameno;

  for(iframe=1;iframe<=nframes;++iframe){
    for(istep=0;istep<nsteps;++istep){

      /* Symplectic leapfrog integrator.  First advance positions half
	 step. */

      for(ipart=0;ipart<Nparts;++ipart){
	parts1[ipart].x=parts[ipart].x+0.5*tau*parts[ipart].px;
	parts1[ipart].y=parts[ipart].y+0.5*tau*parts[ipart].py;
      }

      /* Sort particles by boxes they belong to.  */

      for(i=0;i<nxboxes;++i){
	for(j=0;j<nyboxes;++j){
	  boxes[i][j].nparticles=0;
	}
      }

      /* Next major step, compute forces on particles and advance the
	 momentum a full tau step. */

      for(ipart=0;ipart<Nparts;++ipart){
	x1=parts1[ipart].x;
	y1=parts1[ipart].y;
	j=floor(x1/range);
	j=MAX(j,0);
	j=MIN(j,nxboxes-1);
	k=floor(y1/range);
	k=MAX(k,0);
	k=MIN(k,nyboxes-1);   

	/* (x1,y1) belongs to box (j,k) */

	n=boxes[j][k].nparticles;
	boxes[j][k].particle_indexes[n]=ipart;
	boxes[j][k].nparticles=n+1;
      }

      /* Compute forces on particles at half-stepped position. */

      for(ipart=0;ipart<Nparts;++ipart){
	x1=parts1[ipart].x;
	y1=parts1[ipart].y;

	/* First forces from outside walls */

	fx=(-wf0(-x1/b)+wf0((x1-xright)/b))/b;
	fy=(-wf0(-y1/b)+wf0((y1-ytop)/b))/b;

	/* Now forces from particles in neighboring boxes.  Others are
	   too far away to matter.  */

	dfx=0.0;
	dfy=0.0;

	i=floor(x1/range);
	i=MAX(i,0);
	i=MIN(i,nxboxes-1);
	j=floor(y1/range);
	j=MAX(j,0);
	j=MIN(j,nyboxes-1);  

	/* (x1,y1) belongs to box (i,j).   Add up forces from all
	   particles in all neighboring boxes, but don't include any
	   self-force. */

	nneighbors=boxes[i][j].nneighbors;
	pneighbors=boxes[i][j].pneighbors;
	for(k=0;k<nneighbors;++k){
	  pneighbor=pneighbors[k];
	  nparticles=pneighbor->nparticles;
	  particle_indexes=pneighbor->particle_indexes;
	  for(m=0;m<nparticles;++m){
	    jpart=particle_indexes[m];
	    if(jpart==ipart)continue;
	    dx=x1-parts1[jpart].x;
	    dy=y1-parts1[jpart].y;
	    drsq=dx*dx+dy*dy;
	    xpnt=drsq/asq;
	    if(xpnt>147.0)continue;
	    exp1=exp(-xpnt);
	    if(types[ipart]==types[jpart]){
	      dfx+=dx*exp1;
	      dfy+=dy*exp1;
	    }
	    else{
	      exp2=exp(-0.25*xpnt);
	      expcombo=3.0*exp1-0.5*exp2;
	      dfx+=dx*expcombo;
	      dfy+=dy*expcombo;
	    }
	  }
	}
	fx+=2.0*B*dfx/asq;
	fy+=2.0*B*dfy/asq;

	/* All forces on particle ipart are now computed in (fx,fy).
	   Advance the momentum a full tau step.  */
	  
	parts1[ipart].px=parts[ipart].px+tau*fx;
	parts1[ipart].py=parts[ipart].py+tau*fy;
      }

      /* Now advance the position the remaining half step. */


      for(ipart=0;ipart<Nparts;++ipart){
	parts1[ipart].x+=0.5*tau*parts1[ipart].px;
	parts1[ipart].y+=0.5*tau*parts1[ipart].py;
      }

      /* Copy the advanced state back to original state */

      for(ipart=0;ipart<Nparts;++ipart)parts[ipart]=parts1[ipart];
    }

    /* nsteps intervals tau completed.  Generate a frame.  */

    frameit(iframe,parts,types,Nparts);
    ++frameno;
    print_energy(iframe,parts,types,Nparts);
  }
  exit(0);
}

static void print_energy(int iframe, particle parts[], int types[], 
			 int nparts)
{
  double E,px,py,x,y,xpnt,dx,dy,exp1,exp2,expcombo;
  int ipart,jpart;

  E=0.0;
  for(ipart=0;ipart<nparts;++ipart){

    /* First kinetic energy */

    px=parts[ipart].px;
    py=parts[ipart].py;
    E+=0.5*(px*px+py*py);

    /* Now potential energy from outside walls */

    x=parts[ipart].x;
    y=parts[ipart].y;
    E+=wv0(-x/b)+wv0((x-xright)/b)+wv0(-y/b)+wv0((y-ytop)/b);

     /* Now interparticle potential energy */

    for(jpart=0;jpart<ipart;++jpart){
      dx=x-parts[jpart].x;
      dy=y-parts[jpart].y;
      xpnt=(dx*dx+dy*dy)/asq;
      if(xpnt>147)continue;
      exp1=exp(-xpnt);
      if(types[jpart]==types[ipart]){
	E+=B*exp1;
      }
      else{
	exp2=exp(-0.25*xpnt);
	expcombo=3.0*exp1-2.0*exp2;
	E+=B*expcombo;
      }
    }
  }

  printf("%4d %10.5f\n",iframe,E);
}

static void frameit(int iframe,particle parts[], int types[], int nparts)
{
  static char filename[20];
  static CurvePoint walls[5];
  int i,j;

  sprintf(filename,"frames/%04d.ps",frameno);
  SetGraphicsFileName(filename);
  if(OpenGraphics(15.0,11.0)==1){
    printf("Can't open graphics file %d\n",frameno);
    exit(1);
  }
  NewPage(0);
  SetFiducial(0.5,0.5,6.0);
  SetWindow(0.0,xright,0.0,ytop, 0,0.0,xright, 0,0.0,ytop);
  SetBezier(0);


  walls[0].x=0.0;
  walls[0].y=0.0;
  walls[1].x=xright;
  walls[1].y=0.0;
  walls[2].x=xright;
  walls[2].y=ytop;
  walls[3].x=0.0;
  walls[3].y=ytop;
  walls[4]=walls[0];

  SetLineWidth(10);
  DrawCurve2(walls,5,True,True,True,True,False);

  SetColor(1.0,0.0,0.0);
  for(i=0;i<nparts;++i){
    j=indexes[i];
    if(types[j]==0)SetColor(1.0,0.0,0.0);
    else SetColor(0.0,0.0,1.0);
    DrawBullet(parts[i].x,parts[i].y,0.005);
  }
  EndPage();
  CloseGraphics();
}

static void message_frame()
{
  static char filename[20];

  sprintf(filename,"frames/%04d.ps",frameno);
  SetGraphicsFileName(filename);
  if(OpenGraphics(15.0,11.0)==1){
    printf("Can't open graphics file %d\n",frameno);
    exit(1);
  }
  NewPage(0);
  SetFiducial(0.5,0.5,6.0);
  SetWindow(0.0,xright,0.0,ytop, 0,0.0,xright, 0,0.0,ytop);
  SetBezier(0);
  SetLineWidth(5);
  SetFont(0,0.1);
  DrawText(7.5,7.0,0.0, 2,1,1, False,"Molecules");

  EndPage();
  CloseGraphics();
}

  
